/**
 *  Time-stamp:  <2010-11-27 05:29:11 raskolnikov>
 *
 *  @file        bi_dist_graph.tpp
 *  @author      Juan Pedro Bolivar Puente <raskolnikov@es.gnu.org>
 *  @date        Sat Nov 20 23:35:44 2010
 *
 *  Implementacion.
 */

#ifndef PDP_BI_DIST_GRAPH_T_
#define PDP_BI_DIST_GRAPH_T_

#include <cmath>
#include <stdexcept>
#include <vector>

#include "full_dist_graph.hpp"

namespace pdp
{

namespace detail
{

template <typename T, typename U>
void if_free (T& var, U none)
{
    if (var != none)
    {
	var.Free ();
	var = none; 
    }
}

} /* namespace detail */

template <typename T>
void bi_dist_graph<T>::_free ()
{
    detail::if_free (_comm_row, MPI::COMM_NULL);
    detail::if_free (_comm_col, MPI::COMM_NULL);
    detail::if_free (_type_square, MPI::INT);
    detail::if_free (_type_column, MPI::INT);
}

template <typename T>
void bi_dist_graph<T>::_check (MPI::Intracomm& comm, graph<T>& g)
{
    int    size = comm.Get_size ();
    double size_sqrt = std::sqrt (size);
    int    size_sqrt_floor = size_sqrt;
    
    if (size > (int) g.size () ||
	size_sqrt != static_cast<int> (size_sqrt_floor) ||
	g.size () % size_sqrt_floor != 0)
	throw std::runtime_error ("Wrong number of processes for "
				  "bidimensional partition.");
}

template <typename T>
void bi_dist_graph<T>::scatter (MPI::Intracomm& comm, graph<T>& g, int root)
{
    detail::bcast_graph_size (comm, g, root);
    _check (comm, g);
    
    // Guardamos algunos datos localmente
    _comm         = &comm;
    _default_root = root;
    _data         = g.raw ();
    _size         = g.size ();

    _local_range (comm,
		  _local_row_begin, _local_row_end,
		  _local_col_begin, _local_col_end);
    
    // Dividimos los comunicadores colicales y horiz.
    int rank      = comm.Get_rank ();
    int size      = comm.Get_size ();
    int size_sqrt = std::sqrt (size);
    int side      = _size / size_sqrt;
    
    _comm_row_color = rank % size_sqrt;
    _comm_col_color = rank / size_sqrt;
    
    _comm_row = comm.Split (_comm_row_color, _local_row_begin);
    _comm_col = comm.Split (_comm_col_color, _local_col_begin);

    // Definimos los tipos que vamos a necesitar en las transimisiones
    // sucesivas.
    _type_square = label_traits<T>::mpi_type ().Create_vector (side, side, _size);
    _type_column = label_traits<T>::mpi_type ().Create_vector (side, 1,    _size);
    _type_square.Commit ();
    _type_column.Commit ();

    // Finalmente enviamos la matriz.
    if (rank == root)
    {
	std::vector<MPI::Request> requests;
	requests.reserve (size - 1);
	
	for (int target = 0; target < size; ++target)
	    if (target != root)
	    {
		int col_p = target % size_sqrt;
		int row_p = target / size_sqrt;
		requests.push_back (
		    comm.Isend (_data [row_p * side] + col_p * side, 1,
				_type_square, target, 0));
	    }

	MPI::Request::Waitall (requests.size (), &requests[0]);
    }
    else
	comm.Recv (_data[_local_row_begin] + _local_col_begin,
		   1, _type_square, root, 0);
}

template <typename T>
void bi_dist_graph<T>::bcast_row (size_t row)
{
    int side = _size / _comm_row.Get_size ();
    int root = row / side;
    
    _comm_row.Bcast (_data [row] + side * _comm_row_color, side,
		      label_traits<T>::mpi_type (), root);
}

template <typename T>
void bi_dist_graph<T>::bcast_col (size_t col)
{
    int side = _size / _comm_col.Get_size ();
    int root = col / side;

    _comm_col.Bcast (_data [side * _comm_col_color] + col,
		     1, _type_column, root);
}

template <typename T>
void bi_dist_graph<T>::gather (int root)
{
    if (root < 0)
	root = _default_root;

    int size      = _comm->Get_size ();
    int rank      = _comm->Get_rank ();
    int size_sqrt = _comm_row.Get_size ();
    int side      = _size / size_sqrt;

    if (rank == root)
    {
	std::vector<MPI::Request> requests;
	requests.reserve (size - 1);
	
	for (int from = 0; from < size; ++from)
	    if (from != root)
	    {
		int col_p = from % size_sqrt;
		int row_p = from / size_sqrt;
		int start = row_p * side * _size + col_p * side;
		requests.push_back (
		    _comm->Irecv (*_data + start, 1, _type_square, from, 0));
	    }

	MPI::Request::Waitall (requests.size (), &requests[0]);		
    }
    else
    {
	int start = _local_col_begin + _local_row_begin * _size;
	_comm->Send (*_data + start, 1, _type_square, root, 0);
    }    
}

} /* namespace pdp */

#endif /* PDP_BI_DIST_GRAPH_T_ */

