/**
 *  Time-stamp:  <2011-01-26 18:48:22 raskolnikov>
 *
 *  @file        dist_bnb.tpp
 *  @author      Juan Pedro Bolivar Puente <raskolnikov@es.gnu.org>
 *  @date        Sat Jan 15 14:29:40 2011
 *
 *  Implementacion del branch and bound paralelo.
 */

#ifndef PDP_DIST_BNB_TPP_
#define PDP_DIST_BNB_TPP_

#include <vector>
#include <numeric>
#include <functional>

namespace pdp
{
namespace bnb
{

namespace detail
{

enum message_type
{
    msg_work_request = 0,
    msg_work_arrival,
    msg_baton,
    msg_finish,
    msg_merge
};

} /* namespace detail */

template <typename Node>
void dist_state<Node>::splice_work (stack_type& new_stack)
{
    typename stack_type::iterator it = base::_stack.begin ();
    while (it != base::_stack.end ())
    {
        typename stack_type::iterator spliced = it++;
        new_stack.splice (new_stack.begin (), base::_stack, spliced);
        if (it != base::_stack.end ())
            ++it;
    }
}

template <typename Node>
bool dist_state<Node>::share_work ()
{
    bool done = false;
    int  rank = _comm.Get_rank ();
    int  size = _comm.Get_size ();
        
    if (base::empty ())
    {
        _comm.Send (&rank, 1, MPI::INT, (rank + 1) % size,
                    detail::msg_work_request);
        
        while (base::empty () && !done)
        {
            MPI::Status status;
            _comm.Probe (MPI::ANY_SOURCE, MPI::ANY_TAG, status);
            
            switch (status.Get_tag ())
            {
            case detail::msg_work_request:
                handle_msg_work_request (status);
                break;
                
            case detail::msg_work_arrival:
                handle_msg_work_arrival (status);
                break;

            case detail::msg_baton:
                done = handle_msg_baton (status);
                break;

            case detail::msg_finish:
                done = handle_msg_finish (status);
                break;
                
            default:
                throw std::runtime_error ("Unknown tag.");
            }
        }
    }
    
    if (!done)
    {
        MPI::Status status;
        while (base::size () > 2 &&
               _comm.Iprobe (MPI::ANY_SOURCE,
                             detail::msg_work_request,
                             status))
            handle_msg_work_request (status);
    }
    
    return done;
}

template <typename Node>
void dist_state<Node>::handle_msg_work_request (MPI::Status& status)
{
    const int size = _comm.Get_size ();
    const int rank = _comm.Get_rank ();
    const int next_proc = (rank + 1) % size;
    const int prev_proc = (rank - 1 + size) % size;

    int requester;
    
    _comm.Recv (&requester, 1, MPI::INT, status.Get_source (),
                detail::msg_work_request);

    if (base::empty ())
    {
        _comm.Send (&requester, 1, MPI::INT, next_proc, detail::msg_work_request);

        if (requester == rank)
        {
            _active = false;
            if (_has_baton)
            {
                color baton = _color;  
                if (rank == _root)
                    baton = color_white;

                _comm.Send (&baton, 1, MPI_INT, prev_proc, detail::msg_baton);
                _has_baton = false;
                _color     = color_white;
            }
        }
    }
    else if (base::size () > 2)
    {
        stack_type new_stack;
        splice_work (new_stack);

        typename stack_type::iterator it;
        std::size_t work_size = 0;
        for (it = new_stack.begin (); it != new_stack.end (); ++it)
            work_size += it->mpi_size (_comm);

        std::vector<char> buffer (work_size);
        int position = 0;
        for (it = new_stack.begin (); it != new_stack.end (); ++it)
            it->mpi_pack (&buffer [0], work_size, position, _comm);

        _comm.Send (&buffer[0], position, MPI::PACKED,
                    requester, detail::msg_work_arrival);

        if (rank < requester)
            _color = color_black;
    }
    else
    {
        _comm.Send (&requester, 1, MPI::INT, next_proc, detail::msg_work_request);
    }
}

template <typename Node>
void dist_state<Node>::handle_msg_work_arrival (MPI::Status& status)
{   
    int count = status.Get_count (MPI::PACKED);
    std::vector<char> buffer (count);

    _comm.Recv (&buffer [0], count, MPI::PACKED,
                status.Get_source (), detail::msg_work_arrival);

    int position = 0;
    while (position < count)
        base::push_back (Node::mpi_unpack (&buffer [0], count, position, _comm));

    if (base::empty ())
    {
        const int rank = _comm.Get_rank ();
        const int size = _comm.Get_size ();
        _comm.Send (&rank, 1, MPI::INT, (rank + 1) % size,
                    detail::msg_work_request);
    }
    
    _active = true;
    _color  = color_black;
}

template <typename Node>
bool dist_state<Node>::handle_msg_baton (MPI::Status& status)
{
    const int size      = _comm.Get_size ();
    const int rank      = _comm.Get_rank ();
    const int next_proc = (rank + 1) % size;
    const int prev_proc = (rank - 1 + size) % size;
    
    bool done = false;
    int  msg;
    
    _comm.Recv (&msg, 1, MPI::INT, status.Get_source (), detail::msg_baton);
    
    _has_baton = true;
    if (!_active)
    {
        if (rank == _root && _color == color_white && msg == color_white)
        {
            done = true;
            _comm.Send (&rank, 1, MPI::INT, prev_proc, detail::msg_finish);
            _comm.Recv (&msg,  1, MPI::INT, next_proc, detail::msg_finish);
        }	  
        else
        {
            msg = rank == _root ? color_white : _color;            
            _comm.Send (&msg, 1, MPI::INT, prev_proc, detail::msg_baton);
            _has_baton = false;
            _color = color_white;
        }
    }
    
    return done;
}

template <typename Node>
bool dist_state<Node>::handle_msg_finish (MPI::Status& status)
{
    const int size      = _comm.Get_size ();
    const int rank      = _comm.Get_rank ();
    const int next_proc = (rank + 1) % size;
    const int prev_proc = (rank - 1 + size) % size;

    bool done = true;
    int  msg;
    
    _comm.Recv (&msg,  1, MPI::INT, next_proc, detail::msg_finish);
    _comm.Send (&rank, 1, MPI::INT, prev_proc, detail::msg_finish);

    return done;
}

template <typename Node>
void dist_state<Node>::share_best_cost ()
{
    const int size      = _comm.Get_size ();
    const int rank      = _comm.Get_rank ();
    const int next_proc = (rank + 1) % size;
    const int prev_proc = (rank - 1 + size) % size;
  
    if (_bcast_local_bc && !_wait_bc_return)
    {
        _bc_comm.Send (&_best_cost, 1, node_traits<Node>::mpi_type (),
                       next_proc, rank);
        
        _wait_bc_return = true;
        _bcast_local_bc = false;
    }
  
    MPI::Status status;
    while (_bc_comm.Iprobe (prev_proc, MPI::ANY_TAG, status))
    {
        cost_type new_bc;
        
        _bc_comm.Recv (&new_bc, 1, node_traits<Node>::mpi_type (), prev_proc,
                       MPI::ANY_TAG);
        
        if (new_bc < _best_cost)
            _best_cost = new_bc;		
        
        int msg_source = status.Get_tag ();

        if (msg_source == rank && _bcast_local_bc)
        {
            _bc_comm.Send (&_best_cost, 1, node_traits<Node>::mpi_type (),
                           next_proc, rank);
            
            _wait_bc_return = true;
            _bcast_local_bc = false;
        } 
        else if (msg_source == rank && _bcast_local_bc)
            _wait_bc_return = false;
        else
	{
            _bc_comm.Send (&_best_cost, 1, node_traits<Node>::mpi_type (),
                           next_proc, msg_source);
	}
    }
}

template <typename Node>
void dist_state<Node>::merge_results ()
{
    const int size      = _comm.Get_size ();
    const int rank      = _comm.Get_rank ();

    if (rank == _root)
    {
        cost_type   recv_cost;
        cost_type   best_cost = bnb_cost (base::_best_node);
        int         best_proc = _root;
        MPI::Status status;
        
        for (int i = 0; i < size - 1; ++i)
	{
            _comm.Recv (&recv_cost, 1, node_traits<Node>::mpi_type (),
                        MPI::ANY_SOURCE, detail::msg_merge, status);
            if(recv_cost < best_cost)
            {
                best_cost = recv_cost;
                best_proc = status.Get_source ();
            }
        }

        _comm.Bcast (&best_proc, 1, MPI::INT, _root);

        if (best_proc != _root)
        {
            const int buf_size = base::next ().mpi_size (_comm);
            std::vector<char> buffer (buf_size);
            _comm.Recv (&buffer[0], buf_size, MPI::PACKED, best_proc,
                        detail::msg_merge);
            int pos = 0;
            base::_best_node = Node::mpi_unpack (&buffer[0], buf_size, pos,
                                                _comm);
        }
    }
    else
    {
        int best_cost = bnb_cost (base::_best_node);
        int best_proc;
        
        _comm.Send (&best_cost, 1, node_traits<Node>::mpi_type (),
                    _root, detail::msg_merge);
        _comm.Bcast (&best_proc, 1, MPI::INT, _root);

        if (best_proc == rank)
        {
            const int buf_size = base::_best_node.mpi_size (_comm);
            std::vector<char> buffer (buf_size);
            int pos = 0;
            base::_best_node.mpi_pack (&buffer[0], buf_size, pos, _comm);
            _comm.Send (&buffer[0], buf_size, MPI::PACKED, _root,
                        detail::msg_merge);
        }
    }
}

template <typename Node>
dist_state<Node>::dist_state (const Node& initial, MPI::Intracomm& comm, int root)
    : base (initial)
    , _comm (comm)
    , _bc_comm (comm.Create (comm.Get_group ()))
    , _best_cost (node_traits<Node>::infinity ())
    , _root (root)
    , _active (true)
    , _color (color_black)
    , _has_baton (comm.Get_rank () == root)
    , _bcast_local_bc (true)
    , _wait_bc_return (false)
{
    _best_cost = std::min (base::best_cost (), _best_cost);
}


template <typename Node, typename Context>
Node dist_branch_and_bound (Context& ctx, MPI::Intracomm& comm, int root)
{
    typedef typename node_traits<Node>::cost_type cost_type;
    
    dist_state <Node> stack (node_traits<Node>::initial (ctx), comm, root);

    std::size_t  iterations = 0;
    int          rank = comm.Get_rank ();
    
    if (rank != root)
        stack.pop (); // Borramos el elemento inicial. 

    while (!stack.share_work ())
    {
        ++ iterations;
        stack.pop ();

        cost_type old_cost = stack.best_cost ();
        stack.next ().branch (ctx, stack);

#ifndef PDP_DIST_BNB_NO_SHARE_BEST_COST
        stack.share_best_cost ();
#endif
        
        if (old_cost != stack.best_cost ())
            stack.bound ();
    }

    // HACK: Esto no debería estar aquí
    std::cout << "Proc: " << rank << " Iter: " << iterations << std::endl;
    
    stack.merge_results ();
    
    return stack.best_node ();
}

} /* namespace bnb */
} /* namespace pdp */

#endif /* PDP_BNB_PAR_TPP_ */
