/**
 *  Time-stamp:  <2011-01-25 03:05:32 raskolnikov>
 *
 *  @file        graph.tpp
 *  @author      Juan Pedro Bolivar Puente <raskolnikov@es.gnu.org>
 *  @date        Wed Nov 17 21:41:39 2010
 *
 *  Implementación.
 */

#ifndef PDP_GRAPH_TPP
#define PDP_GRAPH_TPP

#include <stdexcept>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <cstdlib>
#include <iterator>

namespace pdp
{

/**
 * @todo Los operadores << y >> se han implementado siguiendo la
 * implementación del graph.cc original. Sin embargo, una
 * implementación correcta debería satisfacer que los valores escritos
 * y leidos sean equivalentes.
 */
template <typename Label>
std::ostream& operator<< (std::ostream& os, const graph<Label>& g)
{
    for (size_t i = 0; i < g.size (); ++i)
    {
	os << "[" << i << ",*] = ";
	
	for (size_t j = 0; j < g.size (); j++)
	{
	    if (g (i, j) == label_traits<Label>::infinity ()) 
		os << "INF";
	    else  
		os << g (i, j);

	    if (j < g.size () - 1) 
		os << ",";
	    else
		os << std::endl;
	}
    }
    return os;
}

template <typename Label>
std::istream& operator>> (std::istream& is, graph<Label>& g)
{
    size_t graph_size;
    
    if (is >> graph_size)
    {
	g.recreate (graph_size);

#ifdef PDP_GRAPH_FLOYD_FORMAT
	size_t i, j, w;
	while ((is >> i) && (is >> j) && (is >> w))
	    g (i, j) = w;
#else
        std::copy (std::istream_iterator<Label> (is),
                   std::istream_iterator<Label> (), * g.raw ());
#endif
    }
    
    return is;
}

template <typename Label>
graph<Label>::graph (size_t verts)
{
    _allocate (verts);
    _initialise ();
}

template <typename Label>
graph<Label>::graph (const detail::graph_base<Label>& g)
{
    _copy (g);
}

template <typename Label>
graph<Label>::graph (const graph& g)
{
    _copy (g);
}

template <typename Label>
graph<Label>& graph<Label>::operator=  (const detail::graph_base<Label>& g)
{
    const std::size_t s = g.size ();
        
    if (_size != s)
    {
        _liberate ();
        _allocate (s);
    }

    const Label* const* d = g.raw ();
    std::copy (*d, *d + s*s, *_data);
    
    return *this;
}

template <typename Label>
graph<Label>& graph<Label>::operator=  (const graph& g)
{
    return operator= (static_cast<const detail::graph_base<Label>& > (g));
}

#ifdef __GXX_EXPERIMENTAL_CXX0X__

template <typename Label>
graph<Label>::graph (const detail::graph_base<Label>&& g)
{
    _move (g);
}

template <typename Label>
graph<Label>::graph (const graph&& g)
{
    _move (g);
}

template <typename Label>
graph<Label>& graph<Label>::operator=  (const detail::graph_base<Label>&& g)
{
    _move (g);
    return *this;
}

template <typename Label>
graph<Label>& graph<Label>::operator=  (const graph&& g)
{
    _move (g);
    return this;
}

template <typename Label>
void graph<Label>::_move (const detail::graph_base<Label>&& g)
{
    _size = g._size;
    _data = g._data;
    g._size = 0;
    g._data = 0;
}

#endif

template <typename Label>
graph<Label>::~graph ()
{
    _liberate ();
}

template <typename Label>
void graph<Label>::recreate  (size_t verts)
{
    _liberate ();
    _allocate (verts);
    _initialise ();
}

template <typename Label>
void graph<Label>::_allocate (size_t s)
{
    _data = 0;
    _size = s;

    if (s)
    {
	_data  = (Label**) std::malloc (sizeof (Label*)*s + sizeof (Label)*s*s);
	*_data = (Label*) (_data + s);

        for (Label** it = _data + 1; it != _data + s; ++it)
            *it = *(it - 1) + s;
    }
}

template <typename Label>
void graph<Label>::_initialise ()
{
    const std::size_t s = _size;
    
    if (s)
    {
        std::fill (*_data, *_data + s*s, label_traits<Label>::infinity ());
        for (Label* it = *_data; it < *_data + s*s; it += s + 1)
            *it = 0;
    }
}

template <typename Label>
void graph<Label>::_liberate ()
{
    if (_data)
        std::free (_data);
}

template <typename Label>
void graph<Label>::_copy (const detail::graph_base<Label>& g)
{
    const std::size_t s   = g.size ();
    const Label* const* d = g.raw ();
    
    _allocate (s);
    std::copy (*d, *d + s*s, *_data);    
}

} /* namespace pdp */

#endif /* PDP_GRAPH_TPP */
