/**
 *  Time-stamp:  <2011-01-25 03:15:14 raskolnikov>
 *
 *  @file        tsp.tpp
 *  @author      Juan Pedro Bolivar Puente <raskolnikov@es.gnu.org>
 *  @date        Wed Jan 12 12:42:26 2011
 *
 *  Implementacion del viajante de comercio.
 */

#include <cassert>

#ifndef PDP_TSP_T_
#define PDP_TSP_T_

namespace pdp
{
namespace tsp
{

namespace detail
{

template <typename Graph>
bool do_inconsistent (const Graph& g)
{
    for (size_t i = 0; i < g.size (); ++i)
    {
        size_t count = 0;
        for (size_t j = 0; j < g.size (); ++j)
            if (g (i, j) == Graph::infinity ())
                count ++;
        if (count == g.size ())
            return false;
    }

    return true;
}

template <typename Graph>
typename Graph::label_type do_reduce (Graph& g)
{
    const typename Graph::label_type infinity = Graph::infinity ();
    typename Graph::label_type min, cost = 0;
    std::size_t v, w;
    
    for (v = 0; v < g.size (); v++)
    {
        for (w = 0, min = infinity; w < g.size (); w++)
            if (g (v, w) < min && v != w)
                min = g (v, w);
        
        if (min != 0)
        {
            for (w = 0; w < g.size (); w++)
                if (g (v, w) != infinity && v != w)
                    g (v, w) -= min;
            cost += min;
        }
    }

    return cost;
}

} /* namespcae detail */

template <typename Graph>
bool is_inconsistent (const Graph& g)
{
    flipped_graph<typename Graph::value_type> fg (g);
    return
        detail::do_inconsistent (g) &&
        detail::do_inconsistent (fg);
}

template <typename Graph>
typename Graph::value_type reduce (Graph& g)
{
    flipped_graph<typename Graph::value_type> fg (g);
    return
        detail::do_reduce (g) +
        detail::do_reduce (fg);
}

template <typename T>
template <typename Graph>
arc node<T>::choose_arc (const Graph& g) const
{
    for (std::size_t i = 0; i < g.size (); i++)
        if (_incl [i] == null_vertex)
            for (std::size_t j = 0; j < g.size (); j++)
                if (g (i, j) == 0 && i != j)
                    return arc (i, j);

    return null_arc ();
}

template <typename T>
void node<T>::add_arc (arc a)
{
    _incl [a.first] = a.second;

    if (_orig_excl == a.first)
    {
        _orig_excl++;
        for (std::size_t i = 0; i < _dest_excl.size (); i++)
            _dest_excl [i] = null_vertex;
    }
}

template <typename T>
bool node<T>::del_arc (arc a)
{
    if (_orig_excl != a.first)
        return false;
    
    for (std::size_t i = 0; i < _dest_excl.size (); ++i)
        if (_dest_excl [i] == null_vertex) {
            _dest_excl [i] = a.second;
            return true;
        }

    return false;
}

template <typename T>
template <typename Graph>
void node<T>::remove_loops (Graph& g) const
{
    std::size_t cnt, i;
    vertex j;
    
    for (i = 0; i < g.size (); ++i)
        for (cnt = 2, j = _incl [i];
             j != null_vertex && cnt < g.size ();
             cnt++, j = _incl [j])
            g (j, i) = Graph::infinity ();
}

template <typename Graph>
void put_arc (Graph& g, arc a)
{
    vertex j;

    for (j = 0; j < (vertex) g.size (); ++j)
    {
        if (j != a.second)
            g (a.first, j) = Graph::infinity ();
        if (j != a.first)
            g (j, a.second) = Graph::infinity ();
    }
}

template <typename Graph>
void take_arc (Graph& g, arc a)
{
    g (a.first, a.second) = Graph::infinity ();
}

template <typename T>
template <typename Graph>
void node<T>::mark_arcs (Graph& g) const
{
    std::size_t i;
    arc a;

    for (a.first = 0; a.first < (vertex) g.size (); ++a.first)
        if ((a.second = _incl [a.first]) != null_vertex)
            put_arc (g, a);
  
    for (a.first = _orig_excl, i = 0; i < _dest_excl.size (); i++)
        if ((a.second = _dest_excl [i]) != null_vertex)
            take_arc (g, a);
  
    remove_loops (g);
}

template <typename T>
template <typename Graph>
void node<T>::infer_arcs (Graph& g)
{
    bool change;
    arc a;
 
    do {
        change = false;
        for  (std::size_t i = 0; i < g.size (); ++i)
            if (_incl [i] == null_vertex)
            {
                std::size_t cont = 0;
                for (std::size_t j = 0; cont <= 1 && j < g.size (); ++j)
                    if (g (i, j) != Graph::infinity () && i != j)
                    {
                        cont++; 
                        a = arc (i, j);
                    }
                
                if (cont == 1)
                {
                    add_arc (a);
                    put_arc (g, a);
                    remove_loops (g);
                    change = true;
                }
            }
    } while (change);
}

template <typename T>
template <typename Graph1, typename Graph2>
void node<T>::rebuild (Graph1& g1, Graph2& g2)
{
    g2 = g1;
    mark_arcs (g2);
    remove_loops (g2);
    _cost = reduce (g2);
}

template <typename T>
template <typename Graph>
node<T> node<T>::left_child (Graph& g, arc a) const
{
    graph<T> g2 (g);
    node<T> nl  = *this;
    
    nl.add_arc (a);
    nl.mark_arcs (g2);
    nl.infer_arcs (g2);
    nl._cost += reduce (g2);
    
    return nl;
}

template <typename T>
template <typename Graph>
node<T> node<T>::right_child (Graph& g, arc a) const
{
    graph<T> g2 (g);
    node<T> nl  = *this;

    nl.del_arc (a);
    nl.mark_arcs (g2);
    nl.infer_arcs (g2);
    nl._cost += reduce (g2);
    
    return nl;
}

template <typename T>
template <class Graph, class State>
void node<T>::branch (Graph& g, State& s)
{
    graph<T> g2 (g.size ());
    rebuild (g, g2);

    arc a = choose_arc (g2);
    s.push (right_child (g2, a));
    s.push (left_child (g2, a));
}

template <typename T>
bool node<T>::is_solution () const
{
    for (std::size_t i = 0; i < _incl.size (); i++)
        if (_incl [i] == null_vertex)
            return false;
    return true;
}

template <typename T>
std::size_t node<T>::size () const
{
    return std::count (_incl.begin (), _incl.end (), null_vertex);
}

template <typename T>
node<T>::node (std::size_t size)
    : _cost (bnb::node_traits<node>::infinity ())
    , _incl (size, null_vertex)
    , _orig_excl (0)
    , _dest_excl (size > 2 ? size - 2 : 0, null_vertex)
{
}

template <typename T>
node<T>::node (const node& n)
    : _cost (n._cost)
    , _incl (n._incl)
    , _orig_excl (n._orig_excl)
    , _dest_excl (n._dest_excl)
{
}

#ifdef __GXX_EXPERIMENTAL_CXX0X__

template <typename T>
node<T>::node (const node&& n)
    : _cost (n._cost)
    , _incl (std::move (n._incl))
    , _orig_excl (std::move (n._orig_excl))
    , _dest_excl (std::move (n._dest_excl))
{
}

#endif

template <typename T>
std::ostream& operator << (std::ostream& os, const node<T>& n)
{
    std::cout << "ci=" << n._cost << " : ";
    
    for (std::size_t i = 0; i < n._incl.size (); i++)
        if (n._incl [i] != null_vertex)
            std::cout << "<" << i << "," << n._incl [i] << "> ";

    if (n._orig_excl < (vertex) n._incl.size ())
        for (vertex i = 0; i < n._dest_excl [i]; i++)
            if (n._dest_excl [i] != null_vertex)
                std::cout << "!<" << n._orig_excl
                          << ","  << n._dest_excl [i] << "> ";
    std::cout << std::endl;
    return os;
}

template <typename T>
void node<T>::mpi_pack (void* buffer, std::size_t size,
                        int& position, MPI::Comm& c) const
{
    const MPI::Datatype& cost_type   = label_traits<T>::mpi_type ();
    const MPI::Datatype& vertex_type = label_traits<vertex>::mpi_type ();

    vertex problem_size = _incl.size ();
    vertex_type.Pack (&problem_size, 1, buffer, size, position, c);
    
    cost_type.Pack (&_cost, 1, buffer, size, position, c);
    vertex_type.Pack (&_incl [0], _incl.size (), buffer, size, position, c);
    vertex_type.Pack (&_orig_excl, 1, buffer, size, position, c);
    vertex_type.Pack (&_dest_excl [0], _dest_excl.size (),
                      buffer, size, position, c);
}

template <typename T>
node<T> node<T>::mpi_unpack (const void* buffer, std::size_t size,
                             int& position, MPI::Comm& c)
{
    const MPI::Datatype& cost_type   = label_traits<T>::mpi_type ();
    const MPI::Datatype& vertex_type = label_traits<vertex>::mpi_type ();

    vertex problem_size;
    vertex_type.Unpack (buffer, size, &problem_size, 1, position, c);

    node n (problem_size);
    cost_type.Unpack (buffer, size, &n._cost, 1, position, c);
    vertex_type.Unpack (buffer, size, &n._incl [0], n._incl.size (), position, c);
    vertex_type.Unpack (buffer, size, &n._orig_excl, 1, position, c);
    vertex_type.Unpack (buffer, size, &n._dest_excl [0], n._dest_excl.size (),
                        position, c);

    return n;
}

template <typename T>
std::size_t node<T>::mpi_size (const MPI::Comm& c) const
{
    return
        label_traits<T>::mpi_type ().Pack_size (1, c) +
        MPI::INT.Pack_size (1 +                    /* problem_size */
                            _incl.size () +
                            1 +                    /* _orig_excl */
                            _dest_excl.size (),
                            c);
}

} /* namespace tsp */
} /* namespace pdp */

#endif /* PDP_TSP_T_ */
